"use strict";
/**
 * @license
 * Copyright 2017-2019 Balena Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const Promise = require("bluebird");
const parser = require("docker-file-parser");
const jsesc = require("jsesc");
const _ = require("lodash");
const tar = require("tar-stream");
const tar_utils_1 = require("tar-utils");
const generateQemuCopy = (options) => {
    return {
        name: 'COPY',
        args: [options.hostQemuPath, options.containerQemuPath],
    };
};
const processArgString = (argString) => {
    return jsesc(argString, { quotes: 'double' });
};
const transposeArrayRun = (options, command) => {
    const args = command.args.map(processArgString).join(' ');
    return {
        name: 'RUN',
        args: [options.containerQemuPath, '-execve', '/bin/sh', '-c'].concat(args),
    };
};
const transposeStringRun = (options, command) => {
    const processed = processArgString(command.args);
    return {
        name: 'RUN',
        args: [options.containerQemuPath, '-execve', '/bin/sh', '-c'].concat([
            processed,
        ]),
    };
};
const transposeRun = (options, command) => {
    if (_.isArray(command.args)) {
        return transposeArrayRun(options, command);
    }
    return transposeStringRun(options, command);
};
const identity = (options, command) => {
    return command;
};
const commandToTranspose = (command) => {
    if (command.name === 'RUN') {
        return transposeRun;
    }
    return identity;
};
const spaceSeparatedArrayCommands = ['ARG', 'EXPOSE'];
const argsToString = (args, commandName) => {
    if (_.isArray(args)) {
        let ret = '';
        // Handle command meta-arguments (like --from=stage)
        if (args[0] != null && args[0].startsWith('--')) {
            ret += args[0] + ' ';
            args = args.slice(1);
        }
        if (spaceSeparatedArrayCommands.includes(commandName)) {
            return ret + args.join(' ');
        }
        return ret + '["' + args.join('","') + '"]';
    }
    else if (_.isObject(args)) {
        return _.map(args, (value, key) => {
            const escapedValue = JSON.stringify(value);
            return `${key}=${escapedValue}`;
        }).join(' ');
    }
    else {
        return args;
    }
};
const commandsToDockerfile = (commands) => {
    let dockerfile = '';
    commands.map(command => {
        dockerfile += `${command.name} ${argsToString(command.args, command.name)}\n`;
    });
    return dockerfile;
};
/**
 * transpose:
 * Given a string representing a dockerfile, transpose it to use qemu
 * rather than native, to enable emulated builds
 *
 * @param dockerfile A string representing the dockerfile
 * @param options Options to use when doing the transposing
 */
function transpose(dockerfile, options) {
    // parse the Dokerfile
    const commands = parser.parse(dockerfile, { includeComments: false });
    const outCommands = [];
    const copyCommand = generateQemuCopy(options);
    commands.forEach(c => {
        if (c.name === 'FROM') {
            outCommands.push(c);
            outCommands.push(copyCommand);
        }
        else {
            outCommands.push(commandToTranspose(c)(options, c));
        }
    });
    return commandsToDockerfile(outCommands);
}
exports.transpose = transpose;
const getTarEntryHandler = (pack, dockerfileName, opts) => {
    const streamToPromise = require('stream-to-promise');
    return (header, stream, next) => {
        streamToPromise(stream)
            .then((buffer) => {
            const name = tar_utils_1.normalizeTarEntry(header.name);
            if (name === dockerfileName) {
                const newDockerfile = transpose(buffer.toString(), opts);
                pack.entry({ name: dockerfileName }, newDockerfile);
            }
            else {
                if (name === opts.hostQemuPath && opts.qemuFileMode) {
                    header.mode = opts.qemuFileMode;
                }
                pack.entry(header, buffer);
            }
            next();
        })
            .catch(next);
    };
};
/**
 * transposeTarStream: Given a tar stream, this function will extract
 * the files, transpose the Dockerfile using the transpose function,
 * and then re-tar the original contents and the new Dockerfile, and
 * return a new tarStream
 */
function transposeTarStream(tarStream, options, dockerfileName = 'Dockerfile') {
    const extract = tar.extract();
    const pack = tar.pack();
    return new Promise((resolve, reject) => {
        pack.on('error', reject);
        extract.on('error', reject);
        extract.on('entry', getTarEntryHandler(pack, dockerfileName, options));
        extract.on('finish', () => {
            pack.finalize();
            resolve(pack);
        });
        tarStream.pipe(extract);
    });
}
exports.transposeTarStream = transposeTarStream;
/**
 * getBuildThroughStream: Get a through stream, which when piped to will remove all
 * extra output that is added as a result of this module transposing a Dockerfile.
 *
 * This function enables 'silent' emulated builds, with the only difference in output
 * from a native build being that there is an extra COPY step, where the emulator is
 * added to the container
 */
function getBuildThroughStream(opts) {
    const es = require('event-stream');
    // Regex to match against 'Step 1/5:', 'Step 1/5 :' 'Step 1:' 'Step 1 :'
    // and all lower case versions.
    const stepLineRegex = /^(?:step)\s\d+(?:\/\d+)?\s?:/i;
    const isStepLine = (str) => stepLineRegex.test(str);
    // Function to strip the string matched with the regex above
    const stripStepPrefix = (data) => {
        return data.substr(data.indexOf(':') + 1);
    };
    // Regex to match against the type of command, e.g. FROM, RUN, COPY
    const stepCommandRegex = /^\s?(\w+)(:?\s)/i;
    const getStepCommand = (str) => {
        const match = stepCommandRegex.exec(str);
        if (match != null) {
            return match[1].toUpperCase();
        }
        else {
            return '';
        }
    };
    // Regex to remove extra flags which this module adds in
    const replaceRegexString = _.escapeRegExp(`${opts.containerQemuPath} -execve /bin/sh -c `);
    const replaceRegex = new RegExp(replaceRegexString, 'i');
    const replaceQemuLine = (data) => {
        return data.replace(replaceRegex, '');
    };
    return es.pipe(es.mapSync(function (data) {
        data = data.toString();
        if (isStepLine(data) && getStepCommand(stripStepPrefix(data)) === 'RUN') {
            data = replaceQemuLine(data);
        }
        return data;
    }), es.join('\n'));
}
exports.getBuildThroughStream = getBuildThroughStream;

//# sourceMappingURL=index.js.map
